#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

class Sprite(pygame.sprite.Sprite):

    '''
    
    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/spriteExample.gif?raw=true
    
    `Example Sprite code`_.

    .. _Example Sprite code: https://github.com/rik-cross/pygamepal/blob/main/examples/spriteExample.py

    :param str textureURL:  (default = None).
    :param pygame.Texture texture: The sprite texture (default = None).
    :param (int, int) position: The (x, y) position of the top-left of the sprite (default = (0, 0)).
    :param (int, int) size: The (w, h) size of the sprite (default = (0, 0)).
    :param int layer: The draw layer (0 = drawn first, default = 0).
    :param bool scaleTexture: Resize the texture to the sprite size (default = False).
    :param pygamepal.Collider collider: Specify a sprite collider (default = False).
    :param pygamepal.Trigger trigger: Specify a sprite trigger (default = False).
    :param pygame.Color drawColor: The color to draw the sprite outline (default = 'white').
    '''

    def __init__(self,
        textureURL = None,
        texture = None,
        position = (0, 0),
        size = (0, 0),
        layer = 0,
        scaleTexture = False,
        collider = None,
        trigger = None,
        drawColor = 'white'):

        # ensure compatibility with pygame.Sprites
        pygame.sprite.Sprite.__init__(self)
        self.rect = pygame.Rect(position[0], position[1], size[0], size[1])

        # importing here to avoid a circular dependency
        from pygamepal import SpriteImage

        self.position = position

        self.collider = collider
        self.trigger = trigger
        
        self.spriteImage = None

        # use image name to create a texture if defined
        if textureURL is not None:
            texture = pygame.image.load(textureURL)
            texture = texture.convert_alpha()
        
        # create a spriteImage if a texture is specified
        if texture is not None:
            texture = texture.convert_alpha()
            # set the texture size to the sprite size if specified
            if scaleTexture == True and size is not None:
                texture = pygame.transform.scale(texture, size)
            # create the spriteImage
            self.spriteImage = SpriteImage(texture)

        # use the texture size if no size is specified
        if size == (0, 0):
            if texture is not None:
                self.size = texture.get_size()
        else:
            self.size = size
        
        self.layer = layer

        self.currentScene = None

        self.drawColor = drawColor

        # call the user-defined init() method
        self.init()
    
    def _update(self, deltaTime = 1):

        '''
        Update method called once per frame by a parent pygamepal.Scene.
        Users should not call this function.

        :param float deltaTime: The time since the last update call (default = 1).
        '''

        from pygamepal import Trigger, Collider

        # 
        # update the sprite collider
        #

        if self.collider is not None:
            self.collider._sprite = self
            self.collider.update()

        # 
        # update the sprite trigger
        #

        if self.trigger is not None:
            self.trigger._sprite = self
            self.trigger.update()

        #
        # update the sprite image
        #

        if self.spriteImage is not None:
            self.spriteImage.update()
       
        self.update()
    
    def _draw(self, surface):

        '''
        Draw method called once per frame by a parent pygamepal.Scene.
        Users should not call this function.
        
        :param pygame.Surface surface: The surface to draw the sprite to.
        '''

        from pygamepal import Game, drawText, smallFont, DEBUG

        if self.spriteImage is not None:
            self.spriteImage.draw(surface, self.position)

        if DEBUG is True:
            
            # draw sprite size box
            if hasattr(self, 'position') is True and self.position is not None and hasattr(self, 'size') is True and self.size is not None:
                # draw bounding box
                pygame.draw.rect(surface, self.drawColor, (self.position[0], self.position[1], self.size[0], self.size[1]), 1)
                # draw position and size info
                drawText(
                    surface,
                    '[' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.size[0]) + ', ' + str(self.size[1])  + ']',
                    (self.x + self.size[0] + 2, self.y + self.size[1] / 2 - 5),
                    font = smallFont,
                    color = self.drawColor)
            
            # draw sprite collider
            if self.collider is not None:
                self.collider.draw(surface)

            # draw sprite trigger
            if self.trigger is not None:
                self.trigger.draw(surface)
    
    def updateWithoutParentScene(self, deltaTime = 1):

        '''
        Call this draw method once per frame if using a sprite independently of a
        pygamepal.Game or pygamepal.Scene.

        :param float deltaTime: The time since the last update call (default = 1).
        '''

        self._update(deltaTime)

    def drawWithoutParentScene(self, surface):

        '''
        Call this draw method once per frame if using a sprite independently of a
        pygamepal.Game or pygamepal.Scene.

        :param pygame.Surface surface: The surface to draw the sprite to.
        '''
        
        self._draw(surface)

    #
    # user-defined methods
    #

    def init(self):

        '''
        Optional, user-defined init method.
        Called automatically once upon creation if added to a pygamepal.Game or pygamepal.Scene,
        but should be called manually if used independently.
        '''
        
        pass

    def update(self):

        '''
        Optional, user-defined update method.
        Called automatically once per frame if added to a pygamepal.Game or pygamepal.Scene,
        but should be called manually if used independently.
        '''
        
        pass

    def onAddedToScene(self, scene):

        '''
        This method is run once when the sprite is added to a scene.

        :param pygamepal.Scene scene: The scene that the sprite is being added to.
        '''
        
        pass
    
    def onRemovedFromScene(self, scene):

        '''
        This method is run once when the sprite is removed from a scene.

        :param pygamepal.Scene scene: The scene that the sprite is being removed from.
        '''
        
        pass

    def touching(self, otherSprite):

        '''
        Returns True if the sprites position / size intersencts with another sprite.

        :param pygamepal.Sprite otherSprite: The sprite to check against.
        '''
        
        thisRect = pygame.Rect(self.position[0], self.position[1], self.size[0], self.size[1])
        otherRect = pygame.Rect(otherSprite.position[0], otherSprite.position[1], otherSprite.size[0], otherSprite.size[1])
        return thisRect.colliderect(otherRect)
    
    # uses the position and size to return the center
    def getCenter(self):

        '''
        Returns the center (x, y) position of a sprite, using its size.
        '''
        
        return (self.position[0] + self.size[0] / 2, self.position[1] + self.size[1] / 2)

    #
    # properties
    #

    # position, x and y

    @property
    def position(self):
        '''
        Get / set the sprite position, taking any colliders into consideration.
        '''
        return self._position
    
    # setting the position checks colliders
    @position.setter
    def position(self, value):

        from pygamepal import Collider
    
        if hasattr(self, 'collider') is False or self.collider is None:
            # convert the new position to a Vector2
            self._position = pygame.math.Vector2(value[0], value[1])
        
        else:
        
            #
            # check x
            #

            # calculate the travel distance to get the direction
            direction = [value[0] - self._position[0], 0]
            # only check the x movement
            newXPos = [value[0], self.position[1]]
            # get a list of colliding sprites
            collisionList = self.collider._getCollisions(newXPos)
            # find the closes collided sprite
            closestX = None
            for c in collisionList:
                d = c._rect.x - self.collider._rect.x
                if closestX is None or d < closestX:
                    closestX = d
            # move towards the closest collider if one exists
            if closestX is not None:
                if direction[0] > 0:
                    self._position[0] = self._position[0] + closestX - self.collider.size[0]
                elif direction[0] < 0:
                    self._position[0] = self._position[0] + closestX + c._rect[2]
            else:
                self._position[0] = value[0]
            
            #
            # check y
            #
            
            # calculate the travel distance to get the direction
            direction = [0, value[1] - self._position[1]]
            # only check the x movement
            newYPos = [self.position[0], value[1]]
            # get a list of colliding sprites
            collisionList = self.collider._getCollisions(newYPos)
            # find the closes collided sprite
            closestY = None
            for c in collisionList:
                d = c._rect.y - self.collider._rect.y
                if closestY is None or d < closestY:
                    closestY = d
            # move towards the closest collider if one exists
            if closestY is not None:
                if direction[1] > 0:
                    self._position[1] = self._position[1] + closestY - self.collider.size[1]
                elif direction[1] < 0:
                    self._position[1] = self._position[1] + closestY + c._rect[3]
            else:
                self._position[1] = value[1]
            
            # convert the new position to a Vector2
            self._position = pygame.math.Vector2(self._position[0], self._position[1])
        
        # update pygame.Sprite position
        self.rect.x = self.position[0]
        self.rect.y = self.position[1]

    @property
    def x(self):
        '''
        Get / set the sprite x position.
        '''
        return self._position[0]
    
    @x.setter
    def x(self, value):
        self.position = (value, self._position[1])
        self.rect.x = value

    @property
    def y(self):
        '''
        Get / set the sprite y position.
        '''
        return self._position[1]
    
    @y.setter
    def y(self, value):
        self.position = (self._position[0], value)
        self.rect.y = value


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
from math import sin
from random import uniform

class Camera:
    
    '''
    A camera can be used to render any source surface to another destination
    surface, using its attributes (size, position, target position, zoom, etc.).

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/cameraExample.gif?raw=true
    
    `Example Camera code`_.

    .. _Example Camera code: https://github.com/rik-cross/pygamepal/blob/main/examples/cameraExample.py

    :param (int, int) position: the top-left (x, y) coordinate for the camera.
    :param (int, int) size: the size (w, h) of the camera (default = (640, 480), or the parent `Scene` size if used within a `Scene`).
    :param (int, int) target: the coordinate (x, y) of the source surface to draw at the center of the camera (default = (0, 0)).
    :param float lazyFollow: the delay when updating the camera target. 0 = instant snap to target, 1 = no movement (default = 0).
    :param float zoom: the amount to scale the source target (default = 1).
    :param float minZoom: the minimum allowed zoom (default = 0.1).
    :param float maxZoom: the maximum allowed zoom (default = 10).
    :param float lazyZoom: the delay when updating the zoom calue. 0 = instant zoom, 1 = no change in zoom (default = 0).
    :param pygame.Color backgroundColor: the background color (default = 'cornflowerblue').
    :param pygame.Color borderColor: the color of the border (default = 'black').
    :param int borderThickness: the thickness of the border(default = 2).
    :param bool clamp: specifies whether the camera should stay within a specific boundary of the source surface (default = False).
    :param (int, int, int, int) clampRect: the boundary (x, y, w, h) of the source surface that the camera should stay within.
    :param float oscillateSpeed: the speed at which the camera should shake (default = 0.2).
    :param int shakeMagnitude: the magnitude of the camera shake (default = 30).
    :param (float, float) shakeDirection: the (x, y) direction of the shake (default = (1, 0)).
    :param float shakeDampening: the reduction in camera shake magnitude each frame (default = 0.4).
    :param float shakeNoise: the amount of random noise to add to the shake (default = 0.8).
    '''

    def __init__(
        self, position = (0, 0),
        size = (640, 480),
        target = (0, 0),
        lazyFollow = 0,
        zoom = 1, minZoom = 0.1, maxZoom = 10,
        lazyZoom = 0,
        backgroundColor = 'cornflowerblue',
        borderColor='black',
        borderThickness = 2,
        clamp = False,
        clampRect = (0, 0, 1000, 1000),
        oscillateSpeed = 0.2,
        shakeMagnitude = 30,
        shakeDirection = (1, 0),
        shakeDampening = 0.4,
        shakeNoise = 0.8
        ):

        self.position = position
        self.size = size

        # sets the camera target info
        self.target = target
        self._currentTarget = self.target
        self._lazyFollow = lazyFollow

        # sets the zoom info
        self.minZoom = minZoom
        self.maxZoom = maxZoom
        self.zoom = zoom
        self._currentZoom = zoom
        self._lazyZoom = lazyZoom
        
        # background
        self.backgroundColor = backgroundColor

        # border
        self.borderColor = borderColor
        self.borderThickness = borderThickness
        
        # clamp
        self.clamp = clamp
        self.clampRect = clampRect

        # screen shake
        self.oscillateSpeed = oscillateSpeed
        self.shakeMagnitude = shakeMagnitude
        self.shakeDirection = shakeDirection
        self.shakeDampening = shakeDampening
        self.shakeNoise = shakeNoise
        self._shakeTime = 0
        self._shakeCurrent = (0, 0)
        self._shakeCurrentMagnitude = 0

    def update(self, deltaTime = 1):

        '''
        Updates the camera target position and, zoom and shake. This method must be called
        each frame if using a camera in isolation, but the method is called automatically
        for cameras with a parent scene.

        :param float deltaTime: the game time elapsed since the last frame (default = 1).
        '''

        #
        # clamp the camera to the clampRect if required
        #

        if self.clamp:
            # clamp x
            if (self.clampRect[2] - self.clampRect[0]) * self._currentZoom > self.size[0]:
                left = ((self.size[0] / 2) / self._currentZoom) + self.clampRect[0]
                right = ((self.size[0] / 2) / self._currentZoom * -1) + self.clampRect[2]
                self.target = (max(left, min(right, self.target[0])),
                               self.target[1])
            else:
                self.target = (self.clampRect[2] / 2, self.target[1])

            # clamp y
            if (self.clampRect[3] - self.clampRect[1]) * self._currentZoom > self.size[1]:
                top = ((self.size[1] / 2) / self._currentZoom) + self.clampRect[1]
                bottom = ((self.size[1] / 2) / self._currentZoom * -1) + self.clampRect[3]
                self.target = (self.target[0],
                               max(top, min(bottom, self.target[1])))
            else:
                self.target = (self.target[0], self.clampRect[3] / 2)
        
        #
        # update the current target using the target and 'lazy follow' values
        #

        self._currentTarget = (self._currentTarget[0] * self._lazyFollow + self.target[0] * (1 - self._lazyFollow),
                               self._currentTarget[1] * self._lazyFollow + self.target[1] * (1 - self._lazyFollow))
        
        #
        # update screen shake
        #

        # oscillate in a sin wave over time, based on the oscillate speed
        self._shakeTime += deltaTime * self.oscillateSpeed
        shakeOffset = (sin(self._shakeTime), sin(self._shakeTime))
        # add some randomness, based on the noise
        dx = uniform(-1, 1) * self.shakeNoise
        dy = uniform(-1, 1) * self.shakeNoise
        shakeOffset = (shakeOffset[0] + dx, shakeOffset[1] + dy)
        # clamp the offset to the maximum shake offset
        shakeOffset = (max(-1, min(1, shakeOffset[0])), max(-1, min(1, shakeOffset[1])))
        self._shakeCurrent = (shakeOffset[0] * self._shakeCurrentMagnitude * self.shakeDirection[0],
                              shakeOffset[1] * self._shakeCurrentMagnitude * self.shakeDirection[1])
        # reduce the shake magnitude by the dampening amount
        self._shakeCurrentMagnitude = max(0, self._shakeCurrentMagnitude - self.shakeDampening)
        # reset the shake time once the shake is complete
        if self._shakeCurrentMagnitude == 0:
            self._shakeTime = 0

        # update the current zoom amount using the target and 'lazy zoom' values
        self._currentZoom = self._currentZoom * self._lazyZoom + self.zoom * (1 - self._lazyZoom)
                                  
    def draw(self, surface, destSurface):

        '''
        Draws the source surface to the destination surface, using the camera attributes.
        This method must be called each frame if using a camera in isolation, but the method
        is called automatically for cameras with a parent scene.

        :param pygame.Surface surface: the surface to draw.
        :param pygame.Surface destSurface: the surface to draw to.
        '''

        # draw border
        pygame.draw.rect(destSurface, self.borderColor, 
                         (self.position[0] - self.borderThickness, self.position[1] - self.borderThickness, 
                          self.size[0] + self.borderThickness * 2, self.size[1] + self.borderThickness * 2), self.borderThickness, border_radius = 1)
        # ensure that the surface is clipped to the camera dimensions
        destSurface.set_clip((self.position[0], self.position[1], self.size[0], self.size[1]))
        # fill the surface to the background color
        destSurface.fill(self.backgroundColor)

        # blit the (zoomed) surface to the destination, and set the target as the center
        x = 0 - (self.size[0] / 2 - self._currentTarget[0] * self._currentZoom)
        y = 0 - (self.size[1] / 2 - self._currentTarget[1] * self._currentZoom)
        # add screen shake
        x += self._shakeCurrent[0]
        y += self._shakeCurrent[1]

        # draw the surface to the destination using the correct position, size, center and zoom
        destSurface.blit(pygame.transform.scale(surface, (surface.get_width() * self._currentZoom, surface.get_height() * self._currentZoom)), 
                         self.position, 
                         (x, y,
                          self.size[0], self.size[1]))
        # reset surface clipping
        destSurface.set_clip()

    def shake(self, direction = None):

        '''
        Start a camera shake.

        :param tuple(float, float) direction: the (x, y) direction of the shake (default = None, use the stored direction).
        '''

        if direction is not None:
            self.shakeDirection = direction
        self._shakeCurrentMagnitude = self.shakeMagnitude
    
    def setTarget(self, value, instant = False):

        '''
        Set the camera target position. This method permits the instant setting of a position.

        :param bool instant: set the position instantly, ignoring the stored 'lazyFollow' value (default = False).
        '''

        self.target = value
        if instant is True:
            self._currentTarget = value

    def setZoom(self, value, instant = False):

        '''
        Set the camera target zoom value. This method permits the instant setting of the zoom.

        :param bool instant: set the zoom value instantly, ignoring the stored 'lazyZoom' value (default = False).
        '''

        self.zoom = value
        if instant is True:
            self._currentZoom = value

    #
    # properties
    #

    # zoom property, clamped between 0 and 1
    @property
    def zoom(self):
        '''
        Get / set the zoom value (between specified 'minZoom' and 'maxZoom' values).
        '''
        return self._zoom
    
    @zoom.setter
    def zoom(self, value):
        self._zoom = max(self.minZoom, min(self.maxZoom, value))    
        
    @property
    def lazyFollow(self):
        '''
        Get / set the lazy follow value, between 0 and 1.
        '''
        return self._lazyFollow
    
    @lazyFollow.setter
    def lazyFollow(self, value):
        self._lazyFollow = max(0, min(1, value))

    @property
    def lazyZoom(self):
        '''
        Get / set the lazy zoom value, between 0 and 1.
        '''
        return self._lazyZoom
    
    @lazyZoom.setter
    def lazyZoom(self, value):
        self._lazyZoom = max(0, min(1, value))



#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
from .easingFunctions import easeLinear

class Transition:

    '''
    The base class for transitions between 2 surfaces. This class does not need to be used directly.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/transitionExample.gif?raw=true

    `Example Transition code`_.

    .. _Example Transition code: https://github.com/rik-cross/pygamepal/blob/main/transitionExample.py
    
    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):

        self.fromSurface = fromSurface
        self.toSurface = toSurface
        self.duration = duration
        self.easingFunction = easingFunction

        self.easeValue = 0
        self.currentPercentage = 0
        self.finished = False

        # allows adjusting for different durations
        self.durationCoefficient = 100 / self.duration

    def update(self, deltaTime=1):
        perc = self.currentPercentage + (deltaTime * self.durationCoefficient)
        self.currentPercentage = min(perc, 100)
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.currentPercentage == 100:
            self.finished = True

    def draw(self, surface):
        if self.drawMethod is not None:
            self.drawMethod

    def reset(self):
        self.currentPercentage = 0
        self.easeValue = 0
        self.durationCoefficient = 100 / self.duration

class TransitionFade(Transition):

    '''
    Fade between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = easeLinear):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            self.fromSurface.set_alpha((100 - self.easeValue) * 2.55)
            surface.blit(self.fromSurface, (0, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(self.easeValue * 2.55)
            surface.blit(self.toSurface, (0, 0))

    def __repr__(self):
        return 'Transition: Fade between'

class TransitionFadeToBlack(Transition):

    '''
    Fade to black between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)
        self.blacksurface = pygame.Surface((self.fromSurface.get_width(), self.fromSurface.get_height()), pygame.SRCALPHA, 32)
        self.blacksurface.fill('black')

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None and self.easeValue < 50:
            surface.blit(self.fromSurface, (0, 0))
        if self.toSurface is not None and self.easeValue > 50:
            surface.blit(self.toSurface, (0, 0))
        distanceFromMidPoint = abs(50 - self.easeValue)
        percentageToAlpha = 255 - (distanceFromMidPoint * 2.55 * 2)
        self.blacksurface.set_alpha(percentageToAlpha)
        surface.blit(self.blacksurface, (0, 0))

    def __repr__(self):
        return 'Transition: Fade to black'

class TransitionWipeLeft(Transition):

    '''
    Wipe left between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            surface.blit(self.fromSurface, (0, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = (surface.get_width()) - (surface.get_width() / 100 * self.easeValue)
            surface.blit(self.toSurface, (r, 0))

    def __repr__(self):
        return 'Transition: Wipe left'

class TransitionWipeRight(Transition):

    '''
    Wipe right between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            surface.blit(self.fromSurface, (0, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = ((surface.get_width() * -1)) + (surface.get_width() / 100 * self.easeValue)
            surface.blit(self.toSurface, (r, 0))

    def __repr__(self):
        return 'Transition: Wipe right'

class TransitionWipeUp(Transition):

    '''
    Wipe up between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            surface.blit(self.fromSurface, (0, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = (surface.get_height()) - (surface.get_height() / 100 * self.easeValue)
            surface.blit(self.toSurface, (0, r))

    def __repr__(self):
        return 'Transition: Wipe up'

class TransitionWipeDown(Transition):

    '''
    Wipe down between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            surface.blit(self.fromSurface, (0, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = ((surface.get_height() * -1)) + (surface.get_height() / 100 * self.easeValue)
            surface.blit(self.toSurface, (0, r))

    def __repr__(self):
        return 'Transition: Wipe down'

class TransitionMoveLeft(Transition):

    '''
    Move left between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            self.fromSurface.set_alpha(255)
            r = 0 - (surface.get_width() / 100 * self.easeValue)
            surface.blit(self.fromSurface, (r, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = (surface.get_width()) - (surface.get_width() / 100 * self.easeValue)
            surface.blit(self.toSurface, (r, 0))

    def __repr__(self):
        return 'Transition: Move left'

class TransitionMoveRight(Transition):

    '''
    Move right between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            self.fromSurface.set_alpha(255)
            r = (surface.get_width() / 100 * self.easeValue)
            surface.blit(self.fromSurface, (r, 0))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = ((surface.get_width() * -1)) + (surface.get_width() / 100 * self.easeValue)
            surface.blit(self.toSurface, (r, 0))

    def __repr__(self):
        return 'Transition: Move right'

class TransitionMoveUp(Transition):

    '''
    Move up between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            self.fromSurface.set_alpha(255)
            r = 0 - (surface.get_height() / 100 * self.easeValue)
            surface.blit(self.fromSurface, (0, r))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = (surface.get_height()) - (surface.get_height() / 100 * self.easeValue)
            surface.blit(self.toSurface, (0, r))

    def __repr__(self):
        return 'Transition: Move up'

class TransitionMoveDown(Transition):

    '''
    Move down between 2 surfaces.

    :param pygame.Surface from Surface: The surface to transition from.
    :param pygame.Surface toSurface: The surface to transition to.
    :param float duration: The time over which to transition between surfaces.
    :param func easingFunction: The easing function to apply to the transition (default = pygamepal.easeLinear).
    '''

    def __init__(self, fromSurface = None, toSurface = None, duration = 100, easingFunction = None):
        super().__init__(fromSurface, toSurface, duration, easingFunction)

    def draw(self, surface):
        self.easeValue = self.easingFunction(self.currentPercentage)
        if self.fromSurface is not None:
            self.fromSurface.set_alpha(255)
            r = (surface.get_height() / 100 * self.easeValue)
            surface.blit(self.fromSurface, (0, r))
        if self.toSurface is not None:
            self.toSurface.set_alpha(255)
            r = ((surface.get_height() * -1)) + (surface.get_height() / 100 * self.easeValue)
            surface.blit(self.toSurface, (0, r))

    def __repr__(self):
        return 'Transition: Move down'


#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
from pygamepal import Game

class Collider:

    '''
    A collider is an area for which other colliders should not enter.
    A collider can be used independently, or as part of a sprite.

    `Example Collider code`_.

    .. _Example Collider code: https://github.com/rik-cross/pygamepal/blob/main/examples/colliderExample.py

    `Example Sprite Collider code`_.

    .. _Example Sprite Collider code: https://github.com/rik-cross/pygamepal/blob/main/examples/spriteExample.py

    :param:
    '''

    # a static list of all colliders created, to facilitate
    # collision detection for all colliders including those
    # that don't belong to a scene or sprite
    _allColliders = []

    def __init__(self, position = (0, 0),  size = (0, 0), offset = (0, 0), drawColor = 'red'):
        Collider._allColliders.append(self)
        self._scene = None
        self._sprite = None
        self.offset = offset
        self.size = size
        self._rect = pygame.rect.Rect(position[0], position[1], size[0], size[1])
        self.drawColor = drawColor
    
    def update(self):

        '''
        Update method needs to be called for a collider if used independently (i.e. if not as part of a pygamepal.Scene or pygamepal.Sprite).
        '''

        # update position if attached to a sprite
        if self._sprite is not None and self._sprite.position is not None:
            self.x = self._sprite.x + self.offset[0]
            self.y = self._sprite.y + self.offset[1]

    def draw(self, surface):

        '''
        Draw method doesn't need to be called, unless the collider is being used independently and needs to be seen.
        (Note: to see colliders used as part of a pygamepal.Scene or pygamepal.Sprite, set pygamepal.DEGUG = True.)
        '''
        
        from pygamepal import drawText, smallFont
        
        # draw collider bouding box
        pygame.draw.rect(surface, self.drawColor, self._rect, 1)
        # draw size and position info
        drawText(
            surface,
            '[' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.w) + ', ' + str(self.h)  + ']',
            (self.x, self.y + self.h + 2),
            font = smallFont,
            color = self.drawColor)

    def _getCollisions(self, newPosition):

        '''
        Returns a list of all colliders intersecting with newPosition, and using all existing offset and size data.

        :param (int, int) newPosition: The new (x, y) position to check collisions against.
        '''

        # build the list of collisions to return
        collisionList = []

        # return all other colliders...
        for collider in Collider._allColliders:
            if collider is not self:
                # ... in the same scene
                if collider._scene is self._scene:
                    # construct a pygame.Rect object for the new position
                    newPosRect = pygame.Rect(newPosition[0] + self.offset[0], newPosition[1] + self.offset[1], self.size[0], self.size[1])
                    # add collider to list if they would intersect
                    if collider._rect.colliderect(newPosRect):
                        collisionList.append(collider)

        return collisionList

    #
    # properties
    #

    @property
    def x(self):
        '''
        Get / set the x position.
        '''
        return self._rect.x
    
    @x.setter
    def x(self, value):
        # handle collision directly if not part of a scene
        if self._sprite is None:
            newPosition = (value, self.y)
            if len(self._getCollisions(newPosition)) == 0:
                self._rect.x = value
        else:
            self._rect.x = value

    @property
    def y(self):
        '''
        Get / set the y position.
        '''
        return self._rect.y
    
    @y.setter
    def y(self, value):
        # handle collision directly if not part of a scene
        if self._sprite is None:
            newPosition = (self.x, value)
            if len(self._getCollisions(newPosition)) == 0:
                self._rect.y = value
        else:
            self._rect.y = value

    @property
    def w(self):
        '''
        Get / set the width.
        '''
        return self._rect.w
    
    @w.setter
    def w(self, value):
        self._rect.w = value

    @property
    def h(self):
        '''
        Get / set the height.
        '''
        return self._rect.h
    
    @h.setter
    def h(self, value):
        self._rect.h = value


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
import pygamepal

class Game:

    '''
    Create a game without setup or boilerplate code.
    A Game allows you to add scenes, triggers, colliders, buttons and sprites.

    `Example Game code`_.

    .. _Example Game code: https://github.com/rik-cross/pygamepal/blob/main/examples/gameExample.py

    :param (int, int) size: The size of the game window (default = (640, 480)).
    :param str caption: The game window caption (default = '').
    :param int fps: The frames-per-second of the game (default = 60).
    :param bool fullscreen: Fullscreen flag.
    '''

    def __init__(self, size = (640, 480), caption = '', fps = 60, fullscreen = False):
       
        pygame.init()

        self.size = size
        self.caption = caption
        self.fps = fps
        self.fullscreen = fullscreen

        self.input = pygamepal.Input()
        self.currentScene = None
        self.previousScene = None
        self.currentScene = pygamepal.Scene(self)

        # start window in windowed or fullscreen mode
        if self.fullscreen:
            self.screen = pygame.display.set_mode(self.size, pygame.FULLSCREEN)
        else:
            self.screen = pygame.display.set_mode(self.size)

        pygame.display.set_caption(self.caption)
        self.clock = pygame.time.Clock() 

        # total elapsed game time
        self.startTime = pygame.time.get_ticks()
        self.gameTime = self.startTime
        self._running = False
        
        # store events so that they aren't 'consumed' by the class
        self.events = []

        self.init()

    #
    # core methods
    #

    def _update(self):

        '''
        The root update method called automatically once per frame.
        This handles all scenes, etc. and doesn't need to be called by the user.
        '''

        self.previousScene = self.currentScene

        # update clock and calculate delta time
        deltaTime = self.clock.tick(self.fps) / 1000
        # calculate total elapsed time
        self.gameTime = pygame.time.get_ticks() - self.startTime

        # reset events
        self.events = []
        # respond to quit event
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self._running = False
            else:
                self.events.append(event)

        # call user-defined update method
        self.update() 

        # update input
        if self.input is not None:
            self.input.update()

        # update the current scene
        if self.currentScene is not None:
            self.currentScene._update()
        
        # call user-defined active methods
        if self.currentScene is not self.previousScene:
            self.previousScene.onInactive()
            self.currentScene.onActive()

    def _draw(self):

        '''
        The root draw method.
        Does not need to be called by the user.
        '''

        if self.currentScene is not None:
            self.currentScene._draw()
        # call user-defined draw() method
        self.draw()
        pygame.display.flip()

    def run(self):

        '''
        Call to start the game.
        '''

        self._running = True
        while self._running:
            self._update()
            self._draw()
        pygame.quit()

    # call quit() to end the game
    def quit(self):

        '''
        Call to end the game.
        '''

        self._running = False

    #
    # user-defined methods
    #

    def init(self):
        
        '''
        Optional, user-defined method.
        Init() is called once when a game is created.
        '''
        
        pass

    def update(self, deltaTime=1):

        '''
        Optional, user-defined method.
        Called once per frame of the game.

        :param float deltaTime: Time elapsed since last update call.
        '''
        
        pass

    def draw(self):

        '''
        Optional, user-defined method.
        Called once per frame of the game, after update() is called.
        '''

        pass

    #
    # alternative methods to add / remove
    # sprites, triggers, colliders and buttons
    # from the current scene, via the game
    #

    # sprite

    def addSprite(self, sprite):

        '''
        Adds a pygamepal.Sprite to the game.

        :param pygamepal.Sprite sprite: The sprite to add to the game.
        '''
        
        if self.currentScene is not None:
            self.currentScene.addSprite(sprite)
    
    def removeSprite(self, sprite):

        '''
        Removes a pygamepal.Sprite from the game.

        :param pygamepal.Sprite sprite: The sprite to remove from the game.
        '''

        if self.currentScene is not None:
            self.currentScene.removeSprite(sprite)
    
    # trigger

    def addTrigger(self, trigger):

        '''
        Adds a pygamepal.Trigger to the game.

        param pygamepal.Trigger trigger: The trigger to remove from the game.
        '''

        if self.currentScene is not None:
            self.currentScene.addTrigger(trigger)
    
    def removeTrigger(self, trigger):

        '''
        Removes a pygamepal.Trigger from the game.

        :param pygamepal.Trigger trigger: The trigger to remove from the game.
        '''

        if self.currentScene is not None:
            self.currentScene.removeTrigger(trigger)

    # collider

    def addCollider(self, collider):

        '''
        Add a pygamepal.Collider to the game.

        :param pygamepal.Collider collider: The collider to add to the game.
        '''

        if self.currentScene is not None:
            self.currentScene.addCollider(collider)
    
    def removeCollider(self, collider):
        
        '''
        Removes a pygamepal.Collider from the game.

        :param pygamepal.Collider collider: The collider to remove from the game.
        '''
        
        if self.currentScene is not None:
            self.currentScene.removeCollider(collider)
    
    # button

    def addButton(self, button):
        
        '''
        Adds a pygamepal.Button to the game.

        :param pygamepal.Button button: The button to add to the game.
        '''
        
        if self.currentScene is not None:
            self.currentScene.addButton(button)
    
    def removeButton(self, button):
        
        '''
        Removes a pygamepal.Button from the game.

        :param pygamepal.Button button: The button to remove from the game.
        '''
        
        if self.currentScene is not None:
            self.currentScene.removeButton(button)

    #
    # properties
    #

    @property
    def icon(self):
        '''
        Get / set the game icon.
        '''
        return self._icon
    
    @icon.setter
    def icon(self, value):
        self._icon = value
        pygame.display.set_icon(self._icon)



#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

# create a default 'system' font
pygame.font.init()
sysFont = pygame.font.SysFont(None, 24)
smallFont = pygame.font.SysFont(None, 14)

def drawText(surface, text,
             position = [0, 0],
             font = None,
             antialias = True,
             color = 'white', backgroundColor = None,
             centerX = False, centerY = False):
    
    '''
    Draw text.

    :param pygame.Surface surface: The surface to draw to.
    :param str text: The text to draw.
    :param (int, int) position: The (x, y) position to draw on the specified surface (default = (0, 0)).
    :param pygame.Font font: The font to draw the text in (default = pygamepal.sysFont).
    :param bool antialias: Antialias text (default = True).
    :param pygame.Color color: Text color (default = 'white').
    :param pygame.Color backgroundColor: Background text color (default = None).
    :param bool centerX: Center horizontally (default = False).
    :param bool centerY: Center vertically (default = False).
    '''

    # use the default 'system' font if none specified
    if font is None:
        font = sysFont

    # create text surface
    textSurface = font.render(text, antialias, color, backgroundColor)

    # center
    if centerX == True:
        position = (position[0] - textSurface.get_rect().width // 2, position[1])
    if centerY == True:
        position = (position[0], position[1] - textSurface.get_rect().height // 2)

    # draw text to surface
    surface.blit(textSurface, position)


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

class Scene:

    '''
    A pygamepal.Game is made up of one or more scenes, which includes a camera, sprites, triggers and colliders.
    Everything added to a scene is handled automatically, with optional init(), update() and draw() methods to define.
    
    `Example Scene code`_.

    .. _Example Scene code: https://github.com/rik-cross/pygamepal/blob/main/examples/sceneExample.py
    
    :param (int, int) surfaceSize: Restricts the size of the scene surface for improved performance (default = the size of the parent pygamepal.Game window).
    '''

    #
    # static sprite ordering methods
    #

    @staticmethod
    def sortByLayer(s):
        
        '''
        Static method to sort sprites by their layer property.
        '''
        
        return s.layer

    @staticmethod
    def sortByTop(s):
        
        '''
        Static method to sort sprites by their top (y-position).
        '''
        
        if s.position is None:
            return 1
        return s.position[1]

    @staticmethod
    def sortByBottom(s):
        
        '''
        Static method to sort sprites by their bottom position (y-position + height).
        '''
        
        if hasattr(s, 'position') is False or s.position is None or hasattr(s, 'size') is False or s.size is None:
            return 1
        return s.position[1] + s.size[1]
    
    @staticmethod
    def sortByLeft(s):
        
        '''
        Static method to sort sprites by their left property (x-position).
        '''
        
        if s.position is None:
            return 1
        return s.position[0]
    
    @staticmethod
    def sortByRight(s):
        
        '''
        Static method to sort sprites by their right position (x-position + width).
        '''

        if s.position is None or s.size is None:
            return 1
        return s.position[0] + s.size[0]

    #
    # main scene methods
    #

    def __init__(self, game, surfaceSize = None):

        from pygamepal import Camera, Lighting

        # a reference to the main game object
        self.game = game

        if surfaceSize is None:
            self.surfaceSize = game.size
        else:
            self.surfaceSize = surfaceSize

        self.backgroundColor = 'cornflowerblue'
        # frame value is incremented each game tick
        self.frame = 0
        # scene and overlay (UI) surfaces
        self.sceneSurface = pygame.Surface(self.surfaceSize, pygame.SRCALPHA, 32)
        self.overlaySurface = pygame.Surface(game.size, pygame.SRCALPHA, 32)
        # default camera takes up the whole scene, target is the middle of the scene
        self.camera = Camera(position=(0, 0), size=game.size, target=(self.game.size[0] / 2, self.game.size[1] / 2))
        # sort sprites by their z (depth) value
        self.sortKey = Scene.sortByLayer
        # a list of the sprites added to the scene
        self.sprites = []

        self._colliders = []
        self._triggers = []
        self._buttons = []
        
        self.lighting = Lighting(self.surfaceSize, ambientLightLevel = 1)

        # run the user-defined sprite init() method
        self.init()

        # update the camera in case its state has changed
        if self.camera is not None:
            self.camera.update()
    
    def addCollider(self, collider):
        '''
        Adds a collider to the scene.

        :param pygamepal.Collider collider: The collider to add.
        '''
        self._colliders.append(collider)
    def removeCollider(self, collider):
        '''
        Removes a collider from the scene.

        :param pygamepal.Collider collider: The collider to remove.
        '''
        self._colliders.remove(collider)
    def addTrigger(self, trigger):
        '''
        Adds a trigger to the scene.

        :param pygamepal.Trigger trigger: The trigger to add.
        '''
        self._triggers.append(trigger)
    def removeTrigger(self, trigger):
        '''
        Removes a trigger from the scene.

        :param pygamepal.Trigger trigger: The trigger to remove.
        '''
        self._triggers.remove(trigger)
    def addButton(self, button):
        '''
        Adds a button to the scene.

        :param pygamepal.Button button: The button to add.
        '''
        self._buttons.append(button)
    def removeButton(self, button):
        '''
        Removes a button from the scene.

        :param pygamepal.Button button: The button to remove.
        '''
        self._buttons.remove(button)

    def _update(self):

        '''
        Internal update method, called automatically once per frame.
        '''

        from pygamepal import Trigger, Collider

        # update the frame counter
        self.frame += 1
        # update the scene camera
        if self.camera is not None:
            self.camera.update()
        # sort the sprites
        self.sprites.sort(key=self.sortKey)

        #
        # set scene for colliders
        #

        # set any colliders in the scene as owned
        # by the scene
        for collider in self._colliders:
            collider._scene = self
            collider.update()

        # set the sprite colliderr's scene for all sprites
        # in the current scene
        for sprite in self.sprites:
            if hasattr(sprite, 'collider') and sprite.collider is not None:
                sprite.collider._scene = self

        #
        # set scene for triggers
        #

        # set any triggers in the scene as owned
        # by the scene
        for trigger in self._triggers:
            trigger._scene = self
            trigger.update()
        
        # set the sprite trigger's scene for all sprites
        # in the current scene
        for sprite in self.sprites:
            if hasattr(sprite, 'trigger') and sprite.trigger is not None:
                sprite.trigger._scene = self
        
        #
        # set scene for buttons
        #

        # set any triggers in the scene as owned
        # by the scene
        for button in self._buttons:
            button._scene = self
            button.update()

        # update each sprite in the scene
        for s in self.sprites:
            s._update()

        # update the lighting
        self.lighting.update()

        # call the user-defined update() method
        self.update()

    def _draw(self):

        '''
        Internal draw method, called automatically once per frame.
        '''

        from pygamepal import DEBUG

        # clear the surfaces
        self.sceneSurface.fill(self.backgroundColor)
        self.overlaySurface.fill((0, 0, 0, 0))
        # call the user-defined draw() method
        self.draw()

        # draw each scene sprite
        for sprite in self.sprites:
            sprite._draw(self.sceneSurface)

        # draw each button
        for button in self._buttons:
            button.draw(self.overlaySurface)

        if DEBUG is True:

            # draw scene triggers
            for trigger in self._triggers:
                trigger.draw(self.sceneSurface)

            # draw scene colliders
            for collider in self._colliders:
                collider.draw(self.sceneSurface)
        
        # draw the lighting onto the scene screen
        self.lighting.draw(self.sceneSurface)

        # use the camera to draw the scene
        if self.camera is not None:
            self.camera.draw(self.sceneSurface, self.game.screen)
        
        # draw the overlay scene (without the camera)
        self.game.screen.blit(self.overlaySurface, (0, 0))
    
    #
    # user-defined methods, initially empty
    # as they are all optional
    #

    def init(self):

        '''
        Optional scene method called once when created.
        '''
        
        pass

    def update(self):

        '''
        Optional scene method called once per frame.
        '''
        
        pass

    def draw(self):

        '''
        Optional scene method called once per frame.
        '''
        
        pass
    
    def onActive(self):

        '''
        This method is called once when the scene becomes the current scene.
        '''
        
        pass
    
    def onInactive(self):

        '''
        This method is called once when the scene is no longer the current scene.
        '''
        
        pass
    
    #
    # scene helper methods
    #

    def addSprite(self, sprite):

        '''
        Adds a sprite to the scene (and removes from other scenes).

        :param pygamepal.Sprite sprite: The sprite to add.
        '''
        
        # remove from previous scene
        self.removeSprite(sprite)
        # add to new scene
        if sprite not in self.sprites:
            self.sprites.append(sprite)
            sprite.currentScene = self
            sprite.onAddedToScene(self)
    
    # remove a sprite from the scene
    def removeSprite(self, sprite):

        '''
        Removes a sprite from the scene.
        
        :param pygamepal.Sprite sprite: The sprite to remove.
        '''
        
        if self.game.currentScene is not None and sprite.currentScene is not None and sprite in self.currentScene.sprites:
            sprite.onRemovedFromScene(self)
            sprite.currentScene.sprites.remove(sprite)


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

class Animation:

    '''
    Creates new animations for the Animator.
    (Users of the PygamePal library should not need to use this class directly. Instead, it is used by pygamepal.Animator).

    :param Object object: The object to animate a value for.
    :param str attribute: A string representation of the attribute to animate.
    :param float newValue: The new end value to animate to.
    :param float duration: The time to animate from the old to new value.
    :param func(x) easingFunction: The easing function for the animnation (default = pygamepal.easeLinear).
    :param type type: Attribute type (avoids type errors (e.g. if set to int), default = float).
    '''

    def __init__(self, object, attribute, newValue, duration, easingFunction, type):
        
        self.object = object
        self.attribute = attribute
        self.initialValue = getattr(self.object, self.attribute)
        self.newValue = newValue
        self.easingFunction = easingFunction
        self.type = type

        # calculate increment per unit of duration
        increment = (self.newValue - self.initialValue) / duration
        # calculate the amount to change per percentage of animation
        self.percentageIncrement = 100 / ((self.newValue - self.initialValue) / (increment))

        # track the progress of the animation
        self.currentPercentage = 0
        self.complete = False
    
    def update(self, deltaTime = 1):

        '''
        Update must be called once per frame.
        
        :param float deltaTime: Time elapsed since last frame (default = 1).
        '''

        # increment the value (clamped between initial and new values)
        val = self.initialValue + (deltaTime * ((self.newValue - self.initialValue) / 100 * self.easingFunction(self.currentPercentage)))
        if self.newValue > self.initialValue:
            val = min(max(self.initialValue, val), self.newValue)
        if self.newValue < self.initialValue:
            val = min(max(self.newValue, val), self.initialValue)
        
        # set the attribute to the incremented value
        setattr(self.object, self.attribute, self.type(val))

        # set final value if complete
        if self.currentPercentage >= 100:
            setattr(self.object, self.attribute, self.newValue)
            self.complete = True
        
        # increment the percentage complete, clamped to 100
        self.currentPercentage += min(100, self.percentageIncrement)


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

class Input():

    '''
    Handle Pygame input more easily, including key/mouse press, release, long-press and double-press.

    `Example Input code`_.

    .. _Example Input code: https://github.com/rik-cross/pygamepal/blob/main/examples/keyInputExample.py
        
    :param int longPressDuration: The number of milliseconds/frames a key needs to be held to register a long press (default = 60).
    :param int mouseDoublePressTimeout: The maximum number of frames between mouse button presses to register a double-press (default = 30).
    '''

    def __init__(self, longPressDuration = 60, doublePressTimeout = 30):
        
        self.longPressDuration = longPressDuration
        self.doublePressTimeout = doublePressTimeout

        # set current and previous states
        self.currentKeyStates = pygame.key.get_pressed()
        self.previousKeyStates = pygame.key.get_pressed()
        self.currentMouseButtonStates = pygame.mouse.get_pressed()
        self.previousMouseButtonStates = pygame.mouse.get_pressed()

        # store previous mouse position
        self.currentMousePosition = pygame.mouse.get_pos()
        self.previousMousePosition = pygame.mouse.get_pos()

        # set long press durations
        self._keyPressDurations = [0 for _ in range(len(self.currentKeyStates))]
        self._mouseButtonDurations = [0 for _ in range(len(self.currentMouseButtonStates))]

        # store info to determine button double-press
        self._keyTimeSinceLastPressed = [0 for _ in range(len(self.currentKeyStates))]
        self._mouseTimeSinceLastPressed = [0 for _ in range(len(self.currentMouseButtonStates))]
        
        # stores the press state of each button
        # 'none' -> 'single' -> 'double'
        self._keyPressStates = ["none" for _ in range(len(self.currentKeyStates))]
        self._mousePressStates = ["none" for _ in range(len(self.currentMouseButtonStates))]

    def update(self, deltaTime = 1):

        '''
        Call once per frame to update.
        (Note: only needs to be called if using independently (i.e. not as part of a game).)

        :param float deltaTime: Time elapsed since last frame (default = 1).
        '''

        #
        # update key press info
        #

        # update key presses
        self.previousKeyStates = self.currentKeyStates
        self.currentKeyStates = pygame.key.get_pressed()

        # update key press durations
        for i in range(len(self._keyPressDurations)):
            if self.currentKeyStates[i]:
                self._keyPressDurations[i] += deltaTime
            else:
                self._keyPressDurations[i] = 0

        # update mouse button time sinse last pressed
        for i in range(len(self._keyTimeSinceLastPressed)):
            if self.isKeyPressed(i) == True :
                self._keyTimeSinceLastPressed[i] = 0
            else:
                self._keyTimeSinceLastPressed[i] += deltaTime

        # update mouse button press state between
        # 'none', 'single' and 'double'
        for i in range(len(self._keyPressStates)):
            # none -> single
            if self._keyPressStates[i] == "none" and self.isKeyPressed(i):
                self._keyPressStates[i] = "single"
            # single -> double
            elif self._keyPressStates[i] == "single" and self.isKeyPressed(i):
                self._keyPressStates[i] = "double"
            # single -> none
            elif self._keyPressStates[i] == "single" and self._keyTimeSinceLastPressed[i] > self.doublePressTimeout:
                self._keyPressStates[i] = "none"
            # double -> none
            elif self._keyPressStates[i] == "double":
                self._keyPressStates[i] = "none"

        #
        # update mouse button info
        #

        # update mouse presses
        self.previousMouseButtonStates = self.currentMouseButtonStates
        self.currentMouseButtonStates = pygame.mouse.get_pressed()

        # update mouse position
        self.previousMousePosition = self.currentMousePosition
        self.currentMousePosition = pygame.mouse.get_pos()
        
        # update mouse button press durations
        for i in range(len(self._mouseButtonDurations)):
            if self.currentMouseButtonStates[i]:
                self._mouseButtonDurations[i] += deltaTime
            else:
                self._mouseButtonDurations[i] = 0
        
        # update mouse button time sinse last pressed
        for i in range(len(self._mouseTimeSinceLastPressed)):
            if self.isMouseButtonPressed(i) == True :
                self._mouseTimeSinceLastPressed[i] = 0
            else:
                self._mouseTimeSinceLastPressed[i] += deltaTime

        # update mouse button press state between
        # 'none', 'single' and 'double'
        for i in range(len(self._mousePressStates)):
            # none -> single
            if self._mousePressStates[i] == "none" and self.isMouseButtonPressed(i):
                self._mousePressStates[i] = "single"
            # single -> double
            elif self._mousePressStates[i] == "single" and self.isMouseButtonPressed(i):
                self._mousePressStates[i] = "double"
            # single -> none
            elif self._mousePressStates[i] == "single" and self._mouseTimeSinceLastPressed[i] > self.doublePressTimeout:
                self._mousePressStates[i] = "none"
            # double -> none
            elif self._mousePressStates[i] == "double":
                self._mousePressStates[i] = "none"

    #
    # key methods
    #

    def isKeyDown(self, keyCode):

        '''
        Returns true if the key denoted by keyCode is held down during the current frame.

        :param pygame.Key keyCode: The key to check.
        '''
        
        if self.currentKeyStates is None or self.previousKeyStates is None:
            return False
        return self.currentKeyStates[keyCode] == True

    def isKeyPressed(self, keyCode):

        '''
        Returns true if the key denoted by keyCode has been pressed down during the current frame.

        :param pygame.Key keyCode: The key to check.
        '''
        
        if self.currentKeyStates is None or self.previousKeyStates is None:
            return False
        return self.currentKeyStates[keyCode] == True and self.previousKeyStates[keyCode] == False

    def isKeyDoublePressed(self, keyCode):

        '''
        Returns true if the key denoted by keyCode has been double-pressed during the current frame.

        :param pygame.Key keyCode: The key to check.
        '''
        
        return self._keyPressStates[keyCode] == "double"
    
    def isKeyReleased(self, keyCode):

        '''
        Returns true if the key denoted by keyCode has been released during the current frame.

        :param pygame.Key keyCode: The key to check.
        '''
        
        if self.currentKeyStates is None or self.previousKeyStates is None:
            return False
        return self.currentKeyStates[keyCode] == False and self.previousKeyStates[keyCode] == True

    def getKeyDownDuration(self, keyCode):

        '''
        Returns the number of frames a keyCode has been held down.

        :param pygame.Key keyCode: The key to check.
        '''
        
        return self._keyPressDurations[keyCode]

    def isKeyLongDown(self, keyCode):

        '''
        Returns true if the key denoted by keyCode is held down for a long press during the current frame.

        :param pygame.Key keyCode: The key to check.
        '''
        
        return self._keyPressDurations[keyCode] >= self.longPressDuration

    def isKeyLongPressed(self, keyCode):

        '''
        Returns true if the key denoted by keyCode has achieved a long press in the current frame.

        :param pygame.Key keyCode: The key to check.
        '''
        
        return self._keyPressDurations[keyCode] == self.longPressDuration

    def getKeyLongPressPercentage(self, keyCode):

        '''
        Returns keyCode progress towards a long press (0%-100%).

        :param pygame.Key keyCode: The key to check.
        '''
        
        return min(100, self._keyPressDurations[keyCode] / self.longPressDuration * 100)

    #
    # mouse methods
    #

    def getMouseCursorPosition(self):

        '''
        Returns the pygame mouse pointer position.
        '''
        
        return self.currentMousePosition

    def getMouseDirection(self):

        '''
        Returns the (x,y) movement of the mouse.
        e.g:
        - (-1, -1) = North-West movement.
        - (1, 1) = South-East movement.
        '''
        
        direction = (0, 0)
        currentPosition = self.currentMousePosition
        if currentPosition[0] < self.previousMousePosition[0]:
            direction = (-1, direction[1])
        elif currentPosition[0] > self.previousMousePosition[0]:
            direction = (1, direction[1])
        if currentPosition[1] < self.previousMousePosition[1]:
            direction = (direction[0], -1)
        elif currentPosition[1] > self.previousMousePosition[1]:
            direction = (direction[0], 1)
        return direction

    def isMouseButtonDown(self, mouseButton):

        '''
        Returns true if the mouse button specified is held down during the current frame.

        :param int mouseButton: The mouse button to check.
        '''
        
        if self.currentMouseButtonStates is None or self.previousMouseButtonStates is None:
            return False
        return self.currentMouseButtonStates[mouseButton] == True
    
    def isMouseButtonPressed(self, mouseButton):

        '''
        Returns true if the mouse button specified has been pressed in the current frame.

        :param int mouseButton: The mouse button to check.
        '''
        
        if self.currentMouseButtonStates is None or self.previousMouseButtonStates is None:
            return False
        return self.currentMouseButtonStates[mouseButton] == True and self.previousMouseButtonStates[mouseButton] == False

    def isMouseButtonDoublePressed(self, mouseButton):

        '''
        Returns true if the mouse button specified has been double-pressed in the current frame.

        :param int mouseButton: The mouse button to check.
        '''
        
        return self._mousePressStates[mouseButton] == "double"

    def isMouseButtonReleased(self, mouseButton):

        '''
        Returns true if the mouse button specified has been released in the current frame.

        :param int mouseButton: The mouse button to check.
        '''
        
        if self.currentMouseButtonStates is None or self.previousMouseButtonStates is None:
            return False
        return self.currentMouseButtonStates[mouseButton] == False and self.previousMouseButtonStates[mouseButton] == True

    def getMouseButtonDownDuration(self, mouseButton):

        '''
        Returns the number of frames a mouse button has been held down.

        :param int mouseButton: The mouse button to check.
        '''
        
        return self._mouseButtonDurations[mouseButton]

    def isMouseButtonLongDown(self, mouseButton):

        '''
        Returns true if the mouse button specified is held down for a long press during the current frame.

        :param int mouseButton: The mouse button to check.
        '''
        
        return self._mouseButtonDurations[mouseButton] >= self.longPressDuration

    def isMouseButtonLongPressed(self, mouseButton):

        '''
        Returns true if the mouse button specified has achieved a long press in the current frame.

        :param int mouseButton: The mouse button to check.
        '''
        
        return self._mouseButtonDurations[mouseButton] == self.longPressDuration

    def getMouseButtonLongPressPercentage(self, mouseButton):

        '''
        Returns mouse button progress towards a long press (0%-100%).

        :param int mouseButton: The mouse button to check.
        '''
        
        return min(100, self._mouseButtonDurations[mouseButton] / self.longPressDuration * 100)



__version__ = "0.9.1"
__author__ = 'Rik Cross'
__license__ = 'MIT'

import pygame

from .game import *
from .scene import *
from .sprite import *

from .input import *
from .camera import *
from .spriteImage import *
from .collider import *
from .trigger import *
from .button import *
from .particle import *
from .particleEmitter import *
from .transition import *
from .animator import *
from .animation import *
from .lighting import *

from .drawText import *
from .spriteTextureList import *
from .splitTexture import *
from .flatten import *
from .easingFunctions import *

DEBUG = False


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

def splitTexture(texture, newTextureSize):

    '''
    Split a texture into a 2D list of sub-textures, using the given size.
    (Use pygamepal.flatten() to flatten the list returned into a single list.)

    :param pygame.texture texture: The original texture to split.
    :param (int, int) newTextureSize: The size (w, h) of each of the new textures.
    :return list(list(pygame.texture)) newTextures: The 2D list of new textures.
    '''

    # the list of textures to return
    newTextures = []

    # use texture.subsurface to loop through the image
    for row in range(0, texture.get_height(), newTextureSize[1]):
        newRow = []
        for column in range(0, texture.get_width(), newTextureSize[0]):
            # add the cropped texture to the list
            newRow.append(texture.subsurface(
                column, row, newTextureSize[0], newTextureSize[1]))
        newTextures.append(newRow)

    return newTextures


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

from .easingFunctions import easeLinear

class Animator:

    '''
    Changes any object attribute's numerical value over time.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/animatorExample.gif?raw=true

    `Example Animator code`_.

    .. _Example Animator code: https://github.com/rik-cross/pygamepal/blob/main/examples/animatorExample.py
    '''

    def __init__(self):
        self.animations = []

    def addAnimation(self, object, attribute, newValue, duration, easingFunction = easeLinear, type=float):

        '''
        Add a new Animation to the Animator.
        
        :param Object object: The object to animate a value for.
        :param str attribute: A string representation of the attribute to animate.
        :param float newValue: The new end value to animate to.
        :param float duration: The time to animate from the old to new value.
        :param func(x) easingFunction: The easing function for the animnation (default = pygamepal.easeLinear).
        :param type type: Parameter type (avoids type errors (e.g. if set to int), default = float).
        '''

        from .animation import Animation
        # create a new Animation object
        self.animations.append(Animation(object, attribute, newValue, duration, easingFunction, type))

    def update(self, deltaTime = 1):

        '''
        Updating the Animator updates each Animation in the Animator. Must be called once per frame.
        '''

        for a in self.animations:
            # update animation
            a.update(deltaTime)
            # remove if complete
            if a.complete:
                self.animations.remove(a)


#
# PygamePal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame

class Trigger:

    '''
    A trigger defines an area that executes callback functions when colliding with other triggers.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/triggerExample.gif?raw=true
    
    `Example Trigger code`_.

    .. _Example Trigger code: https://github.com/rik-cross/pygamepal/blob/main/examples/triggerExample.py

    :param (int, int) position: The (x, y) top-left position of the trigger (default = (0, 0)).
    :param (int, int) size: The (w, h) size of the trigger, in pixels (default = (10, 10)).
    :param (int, int) offset: The (x, y) offset from the parent pygamepal.Sprite position (default = (0, 0)).
    :param func onEnter: A function called once when two triggers first collide (default = None).
    :param func onCollide: A function called on every frame that two triggers collide (default = None).
    :param func onExit: A function called once when two triggers no longer collide (default = None).
    :param pygame.Color drawColor: The color to draw the trigger (default = 'yellow').
    '''
    
    # a static list of all triggers created, to facilitate
    # collision detection for all triggers including those
    # that don't belong to a scene or sprite
    _allTriggers = []

    def __init__(
        self,
        position = (0, 0),
        size = (10, 10),
        offset = (0, 0),
        onEnter = None,
        onCollide = None,
        onExit = None,
        drawColor = 'yellow'):
        
        # add the trigger to the static list
        Trigger._allTriggers.append(self)

        # stores other triggers currently colliding with this one
        self._collidedTriggers = []
        
        # parent scene and sprite
        self._scene = None
        self._sprite = None
        
        # the trigger size and position are stored as a pygame.Rect
        self.offset = offset
        # position is adjusted and offset from the parent sprite if appropriate
        self._rect = pygame.rect.Rect(position[0], position[1], size[0], size[1])
        
        # set trigger collision callbacks
        self.onCollide = onCollide
        self.onEnter = onEnter
        self.onExit = onExit
        
        self.drawColor = drawColor
        # the color changes if a trigger is activated
        self._currentColor = drawColor

    def update(self, deltaTime = 1):

        '''
        Update must be called once per frame if used independently.
        (Note: does not need to be called if used as part of a pygamepal.Game or pygamepal.Scene)

        :param float deltaTime: Time elapsed since last game tick (default = 1).
        '''

        #
        # update trigger position if attached to a parent sprite
        #

        if self._sprite is not None and self._sprite.position is not None:
            self.x = self._sprite.x + self.offset[0]
            self.y = self._sprite.y + self.offset[1]

        #
        # process this trigger with respect to all other triggers
        #

        # check against every created trigger
        for t in Trigger._allTriggers:
            # don't collide trigger with itself
            if t is not self:
                # only collide triggers in the same scene
                if self._scene == t._scene:
                    # if triggers collide
                    if self._rect.colliderect(t._rect):

                        if t not in self._collidedTriggers:
                            # onEnter
                            if self.onEnter is not None:
                                self.onEnter(self, t)
                            self._collidedTriggers.append(t)
                        # onCollide
                        if self.onCollide is not None:
                            self.onCollide(self, t)
                            
                    # if no trigger collision
                    else:
                        if t in self._collidedTriggers:
                            # onExit
                            if self.onExit is not None:
                                self.onExit(self, t)
                            self._collidedTriggers.remove(t)
        
        # change the trigger color if it has been activated
        if len(self._collidedTriggers) > 0:
            self._currentColor = 'green'
        else:
            self._currentColor = self.drawColor
                
    def draw(self, surface):

        '''
        Draw can be called to draw a trigger when used independently.
        (Note: this method is called automatically when pygamepal.DEBUG is True,
        when used as part of a pygamepal.Game or pygamepal.Scene)

        :param pygame.Surface surface: The surface to draw to.
        '''

        from pygamepal import drawText, smallFont
        
        # draw trigger bouding box
        pygame.draw.rect(surface, self._currentColor, self._rect, 1)
        # draw size and position info
        drawText(
            surface,
            '[' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.w) + ', ' + str(self.h)  + ']',
            (self.x, self.y - 12),
            font = smallFont,
            color = self._currentColor)
        
    #
    # properties
    #

    @property
    def x(self):
        '''
        Get / set the trigger x position.
        '''
        return self._rect.x
    
    @x.setter
    def x(self, value):
        self._rect.x = value

    @property
    def y(self):
        '''
        Get / set the trigger y position.
        '''
        return self._rect.y
    
    @y.setter
    def y(self, value):
        self._rect.y = value

    @property
    def w(self):
        '''
        Get / set the trigger width.
        '''
        return self._rect.w
    
    @w.setter
    def w(self, value):
        self._rect.w = value

    @property
    def h(self):
        '''
        Get / set the trigger height.
        '''
        return self._rect.h
    
    @h.setter
    def h(self, value):
        self._rect.h = value


#
# pygamepal, by Rik Cross
#  -- homepage: github.com/rik-cross/pygamepal
#  -- MIT licenced, free to use, modify and distribute
#  -- run 'pip install pygamepal' to use
#

import pygame
from .drawText import drawText

class Button:

    '''
    Used to provide input, and can be controlled by mouse and/or keypress.

    .. image:: https://github.com/rik-cross/pygamepal/blob/main/examples/gifs/buttonExample.gif?raw=true

    `Example Button code`_.

    .. _Example Button code: https://github.com/rik-cross/pygamepal/blob/main/examples/buttonExample.py

    :param pygamepal.Input input: The input object used to control the button.
    :param (int, int) position: The (x, y) position of the button on the drawn surface (default = (0, 0)).
    :param str text: Text to display on the button (default = None).
    :param pygame.Color foregroundColor: Foreground color, used to draw the button text and border (default = 'white).
    :param pygame.Color backgroundColor: Background color, used to draw the button background (default = 'black).
    :param int borderWidth: Width of the button border (default = 1).
    :param pygame.Color borderColor: Color of the button border (default = 'white').
    :param pygame.Texture image: Button background image (default = None).
    :param function(pygamepal.Button) onHighlighted: A function called when the mouse position intersects with the button position (default = None).
    :param function(pygamepal.Button) onSelected: A function called when the button is clicked, or the relevant button pressed (default = None).
    :param function(pygamepal.Button) updateMethod: Used to override the default update method of a button (default = None).
    :param function(pygamepal.Button, pygame.Surface) drawMethod: Used to override the default draw method of a button (default = None).
    :param pygame.Key keyCode: The keypress associated with selecting the button (default = None).
    '''

    def __init__(self,
        # input is not optional
        input,
        position = (0, 0), size = (100, 50),
        text = None,
        foregroundColor = 'white', backgroundColor = 'black',
        borderWidth = 1,
        borderColor = 'white',
        image = None,
        # this method called when highlighted
        onHighlighted = None,
        # this method is called when selected
        onSelected = None,
        # updateMethod and drawMethod give the ability
        # to override default button befaviour
        updateMethod = None,
        drawMethod = None,
        # a keycode can also be associated with a button
        # (only works if pygamepal.input is specified)
        keyCode = None
        ):
        
        self._scene = None

        self.input = input
        self.position = position
        self.size = size
        self.text = text
        
        if image is not None:
            self.image = pygame.image.load(image).convert_alpha()
        else:
            self.image = image

        self.foregroundColor = foregroundColor
        self.backgroundColor = backgroundColor
        self.borderWidth = borderWidth
        self.borderColor = borderColor
        self.onHighlighted = onHighlighted
        self.onSelected = onSelected
        self.updateMethod = updateMethod
        self.drawMethod = drawMethod
        self.keyCode = keyCode

        self._isHighlighted = False
        self._isSelected = False
    
    def update(self):

        '''
        Must be called once per frame if the button is not used as part of a pygamepal.Scene.
        '''

        if self.input is None:
            return

        if self.updateMethod is not None:
            self.updateMethod(self)
            return

        self.setHighlighted()
        self.setSelected()

    def draw(self, surface):

        '''
        Must be called once per frame if the button is not being used as part of a pygamepal.Scene.

        :param pygame.Surface surface: The surface to draw to.
        '''

        if self.drawMethod is not None:
            self.drawMethod(self, surface)
            return

        self.drawBackground(surface)
        self.drawImage(surface)
        self.drawText(surface)
        self.drawBorder(surface)

    #
    # draw helper methods
    # (useful for when providing a draw callback)
    #

    def setHighlighted(self):

        '''
        Should be called within a custom button.updateMethod in order to update the button 'highlighted' status.
        A button is highlighted if the mouse is within the button bounds.
        '''
        
        self._isHighlighted = False
        cursor = self.input.getMouseCursorPosition()
        if self._isWithinBounds(cursor, (self.position[0], self.position[1], self.size[0], self.size[1])):
            self._isHighlighted = True
            if self.onHighlighted is not None:
                self.onHighlighted(self)

    def setSelected(self):

        '''
        Should be called within a custom button.updateMethod in order to update the button 'selected' status.
        A button is selected if either the keyCode is pressed, or if the mouse button is pressed within the button bounds.
        '''

        self._isSelected = False
        cursor = self.input.getMouseCursorPosition()
        if (self.keyCode is not None and self.input.isKeyPressed(self.keyCode)) or \
            (self._isWithinBounds(cursor, (self.position[0], self.position[1], self.size[0], self.size[1])) and \
            self.input.isMouseButtonPressed(0)):
            self._isSelected = True
            if self.onSelected is not None:
                self.onSelected(self)

    def _isWithinBounds(self, position, rect):
        
        '''
        returns True if an (x, y) 'position' (usually the mouse position) is inside the (x, y, w, h) 'rect' (usually the button).
        
        :param (int, int) position: The (x, y) position to test.
        :param (int, int, int, int): The (x, y, w, h) rectangle to test against the position.
        '''
        
        return position[0] >= rect[0] and position[0] <= rect[0] + rect[2] and \
            position[1] >= rect[1] and position[1] <= rect[1] + rect[3]
    
    def drawBackground(self, surface):

        '''
        Draws the button background. Can be called from a custom button.drawMethod function.

        :param pygame.Surface surface: The surface to draw to.
        '''
        
        pygame.draw.rect(surface,
                         self.backgroundColor,
                         (self.position[0], self.position[1], self.size[0], self.size[1]))
    
    def drawImage(self, surface):

        '''
        Draws the button image. Can be called from a custom button.drawMethod function.

        :param pygame.Surface surface: The surface to draw to.
        '''
        
        if self.image is not None:
            surface.blit(self.image, self.position)

    def drawText(self, surface):
        

        '''
        Draws the button text. Can be called from a custom button.drawMethod function.

        :param pygame.Surface surface: The surface to draw to.
        '''

        drawText(surface, text = self.text,
                 position = (self.position[0] + self.size[0] / 2, self.position[1] + self.size[1] / 2),
                 color = self.foregroundColor, centerX = True, centerY = True)

    def drawBorder(self, surface):

        '''
        Draws the button border. Can be called from a custom button.drawMethod function.

        :param pygame.Surface surface: The surface to draw to.
        '''
        
        if self.borderWidth > 0:
            pygame.draw.rect(surface,
                             self.foregroundColor,
                             (self.position[0], self.position[1], self.size[0], self.size[1]),
                             width = 1)
    
    #
    # properties
    #

    # True when the button is hoghlighted
    @property
    def isHighlighted(self):
        '''
        Returns True is the mouse position intersents with the button.
        '''
        return self._isHighlighted
    
    # True when the button is selected
    @property
    def isSelected(self):
        '''
        Returns True is the button keyCode is pressed, or if the button is clicked.
        '''
        return self._isSelected   
